from uuid import UUID
from datetime import datetime, timedelta

import fastapi as fa
import fastapi.security as fs
from passlib.context import CryptContext
from jose import jwt, JWTError

from db.repositories import UserRepository, GroupRepository
from .schemas import TokenData, LoginRequestModel, UserGetSchema
from .const import (
    SECRET_KEY, ACCESS_TOKEN_EXPIRE_MINUTES, REFRESH_TOKEN_EXPIRE_HOURS, ALGORITHM
)
from exceptions import common as common_exc, http as http_exc
from db.models import Group
from ..group.schemas import GroupIdSchema


pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')

oauth2_scheme = fs.OAuth2PasswordBearer(tokenUrl='/api/auth/login')


class TokenService:
    def create_token(self, data: TokenData, expires_delta: timedelta) -> str:
        payload = {
            'sub': str(data.sub),
            'role': data.role,
            'group': data.group,
            'exp': datetime.utcnow() + expires_delta,
        }
        print(payload)
        return jwt.encode(payload, SECRET_KEY, algorithm=ALGORITHM)

    def create_tokens(self, data: TokenData):
        access = self.create_token(
            data, expires_delta=timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
        )
        refresh = self.create_token(
            data, expires_delta=timedelta(hours=REFRESH_TOKEN_EXPIRE_HOURS)
        )

        return {"access_token": access, "refresh_token": refresh}

    def decode_token(self, token: str) -> TokenData:
        credentials_exception = fa.HTTPException(
            status_code=fa.status.HTTP_401_UNAUTHORIZED,
            detail="Could not validate credentials",
            headers={"WWW-Authenticate": "Bearer"},
        )

        try:
            decoded_token_payload = jwt.decode(token, SECRET_KEY, algorithms=[ALGORITHM])
            user_id = decoded_token_payload.get('sub')
            if user_id is None:
                raise credentials_exception

        except JWTError:
            raise credentials_exception

        return TokenData(**decoded_token_payload)

    def refresh_token(self, refresh_token: str):
        token_data = self.decode_token(refresh_token)
        access = self.create_token(token_data, timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES))
        return {'access_token': access, 'refresh_token': refresh_token}


token_service = TokenService()


class AuthService:
    def __init__(self) -> None:
        self.user_repo = UserRepository()

    async def verify_password(self, plain_password: str, hashed_password: str) -> bool:
        return pwd_context.verify(plain_password, hashed_password)

    async def authenticate_user(self, username: str, password: str):
        users = await self.user_repo.get_list(username=username)

        if users and await self.verify_password(password, users[0].password):
            return users[0]

    async def login(self, data: LoginRequestModel):
        user = await self.authenticate_user(username=data.username, password=data.password)

        if not user:
            raise fa.HTTPException(
                status_code=fa.status.HTTP_404_NOT_FOUND,
                detail="Incorrect username or password",
                headers={"WWW-Authenticate": "Bearer"},
            )
        groups = await user.group.all()
        group_uuids = [str(group.id) for group in groups]

        return token_service.create_tokens(
            TokenData(sub=user.id, role=user.role, group=group_uuids)
        )

    async def get_current_user(self, token: str = fa.Security(oauth2_scheme)):
        token_data = token_service.decode_token(token)
        user = await self.user_repo.get(id=token_data.sub)
        return user


auth_service = AuthService()


class UserService:
    user_repository = UserRepository()
    group_repository = GroupRepository()

    async def get_user_list(self, **kwargs):
        users = await self.user_repository.get_list(**kwargs)
        return [UserGetSchema.model_validate(
            {
                'id': user.id,
                'name': user.name,
                'surname': user.surname,
                'email': user.email,
                'role': user.role,
                'avatar': user.avatar,
                'group': await user.group.all().values_list('id', flat=True)
            }
        ) for user in users]

    async def get_user(self, id: UUID):
        user = await self.user_repository.get(id)
        return UserGetSchema.model_validate(user)

    async def create_user(self, **kwargs):
        user = await self.user_repository.create(
            username=kwargs.get('username'),
            password=kwargs.get('password'),
            name=kwargs.get('name'),
            surname=kwargs.get('surname'),
            email=kwargs.get('email'),
            role=kwargs.get('role'),
            avatar=kwargs.get('avatar')
        )
        group_uuids = kwargs.get('group', [])
        groups = await Group.filter(id__in=group_uuids)
        await user.group.add(*groups)
        groups = await user.group.all()
        user = {
            'id': user.id,
            'name': user.name,
            'surname': user.surname,
            'email': user.email,
            'role': user.role,
            'avatar': user.avatar,
            'group': [group.id for group in groups]
        }
        return UserGetSchema.model_validate(user)

    async def update_user(self, id: UUID, **kwargs):
        user = await self.user_repository.update(id, **kwargs)
        return UserGetSchema.model_validate(user)

    async def delete_user(self, id: UUID):
        return await self.user_repository.delete(id)
