import math

import tortoise
from tortoise.expressions import Q
from passlib.context import CryptContext

from .models import (
    Status, Type, Task, User, Comment, Category, Group, Scopes
)
from exceptions import common as common_exc


class BaseRepository:
    model: tortoise.Model

    async def get_list(self, **kwargs) -> list[tortoise.Model]:
        return await self.model.filter(**kwargs)

    async def get(self, id: int) -> tortoise.Model:
        try:
            return await self.model.get(id=id)

        except tortoise.exceptions.DoesNotExist as e:
            raise common_exc.NotFoundException(str(e))

    async def create(self, **kwargs) -> tortoise.Model:
        try:
            return await self.model.create(**kwargs)

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.CreateException(str(e))

    async def update(self, id: int, **kwargs) -> tortoise.Model:
        try:
            instance = await self.get(id=id)
            await instance.update_from_dict(kwargs).save()

            return instance

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.UpdateException(str(e))

    async def delete(self, id: int) -> None:
        try:
            instance = await self.get(id=id)
            await instance.delete()

        except tortoise.exceptions.IntegrityError as e:
            raise common_exc.DeleteException(str(e))


class StatusRepository(BaseRepository):
    model = Status


class TypeRepository(BaseRepository):
    model = Type


class TaskRepository(BaseRepository):
    model = Task

    async def get_list(
        self,
        title: str | None = None,
        status: str | None = None,
        category: str | None = None,
        priority: str | None = None,
        sort_by: str | None = None,
        page: int | None = None,
        size: int | None = None
    ):
        query = self.model.all()

        offset_min = (page - 1) * size
        offset_max = page * size

        if title is not None:
            query = query.filter(title__icontains=title)

        if status is not None:
            query = query.filter(status__name__icontains=status)

        if category is not None:
            query = query.filter(category__name__icontains=category)

        if priority is not None:
            query = query.filter(priority=priority)

        if sort_by is not None:
            query = query.order_by(sort_by)

        data = await query
        result = {
            "page": page,
            "size": size,
            "total": math.ceil(len(data) / size) - 1,
            "result": data[offset_min:offset_max]
        }
        return result

    async def search(self, keywords: str):
            query = self.model
            search_terms = keywords.split()

            conditions = [Q(title__icontains=term) | Q(description__icontains=term) for term in search_terms]

            combined_condition = conditions.pop()

            for condition in conditions:
                combined_condition |= condition

            query = query.filter(combined_condition)

            return await query.all()


class UserRepository(BaseRepository):
    model = User

    @staticmethod
    def get_password_hash(password):
        pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        return pwd_context.hash(password)

    async def create(self, **kwargs):
        kwargs['password'] = self.get_password_hash(kwargs['password'])
        return await super().create(**kwargs)


class CommentRepository(BaseRepository):
    model = Comment


class CategoryRepository(BaseRepository):
    model = Category


class GroupRepository(BaseRepository):
    model = Group


class ScopeRepository(BaseRepository):
    model = Scopes
